import numpy as np
from logging import getLogger

log = getLogger(__file__)

def mutual_information_feature_select(x: np.ndarray,
    y: np.ndarray,
    num_features_or_beta: int,
    max_iter: int = 25,
    lower_beta: float = 0.0,
    upper_beta: float = 1.0,
    ):
    if num_features_or_beta < 0.0:
        raise ValueError()
    elif isinstance(num_features_or_beta, float) or num_features_or_beta == 0:
        return mutual_information_feature_select_by_beta(x, y, num_features_or_beta)
    else:
        return mutual_information_feature_select_by_num_features(x, y, num_features_or_beta, max_iter, lower_beta, upper_beta)

def mutual_information_feature_select_by_num_features(
    x: np.ndarray,
    y: np.ndarray,
    num_features: int,
    max_iter: int,
    lower_beta: float,
    upper_beta: float,
):
    assert num_features > 0
    assert max_iter > 0

    num_found_features = 0

    iter = 0
    while iter < max_iter:
        iter += 1
        
        new_beta = 0.5 * (lower_beta + upper_beta)
        log.info(f"Iteration {iter}/{max_iter}.")
        log.info(f"Guessing beta to be {new_beta}.")
        features = mutual_information_feature_select_by_beta(x, y, new_beta)
        num_found_features = len(features)
        
        log.info(f"Found {num_found_features} features and need {num_features}.")

        if num_found_features < num_features:
            upper_beta = new_beta
        elif num_found_features > num_features:
            lower_beta = new_beta
            if lower_beta > (0.95 * upper_beta):
                upper_beta *= 2
        else:
            return features, new_beta
        
    if iter >= max_iter:
        return features[:num_features], np.nan


def mutual_information_feature_select_by_beta(x: np.ndarray, y: np.ndarray, beta: float = 0.04):
    assert beta >= 0.0
    # assert beta <= 1.0
    
    num_features = x.shape[1]

    x_bin_edges_list = [make_clean_bins_from_data(x_col) for x_col in x[:,]]
    y_bin_edges = make_clean_bins_from_data(y)

    x_probs_list = [
        calculate_probabilities_with_nans(x_col, bin_edges)
        for x_col, bin_edges in zip(x.T, x_bin_edges_list)
    ]
    y_probs = calculate_probabilities_with_nans(y, y_bin_edges)

    I_xx = np.zeros((num_features, num_features))
    I_xy = np.zeros(num_features)

    for ix1, (x1_data, x1_bin_edges, x1_probs) in enumerate(
        zip(x.T, x_bin_edges_list, x_probs_list)
    ):
        # fill I_xx
        for ix2, (x2_data, x2_bin_edges, x2_probs) in enumerate(
            zip(x.T, x_bin_edges_list, x_probs_list)
        ):
            x1x2_probs = calculate_2d_probabilities_with_nans(
                x1_data, x2_data, x1_bin_edges, x2_bin_edges
            )
            mi = mutual_information(x1_probs, x2_probs, x1x2_probs)
            I_xx[ix1, ix2] = mi

        # fill I_y
        xy_probs = calculate_2d_probabilities_with_nans(
            x1_data, y, x1_bin_edges, y_bin_edges
        )
        mi = mutual_information(x1_probs, y_probs, xy_probs)
        I_xy[ix1] = mi

    features = joint_mutual_information_feature_select(I_xx, I_xy, beta)
    return features


def joint_mutual_information_feature_select(I_xx, I_xy, beta):
    selected_features = []
    num_features = len(I_xx)

    while len(selected_features) < num_features:
        best_feature_idx = I_xy.argmax()
        best_mi_to_target = I_xy[best_feature_idx]

        if best_mi_to_target < 0:
            break

        selected_features.append(best_feature_idx)

        selected_mi_to_other_features = I_xx[:, best_feature_idx]

        I_xy -= beta * selected_mi_to_other_features
        I_xy[best_feature_idx] = np.NINF

    return np.array(selected_features)


def calculate_probabilities_with_nans(data, bins):
    nan_mask = np.isnan(data)
    clean_data = data[~nan_mask]

    counts, _ = np.histogram(clean_data, bins)

    nan_counts = nan_mask.sum()

    master_counts = np.empty((counts.shape[0] + 1,), dtype=int)
    master_counts[:-1] = counts
    master_counts[-1] = nan_counts

    master_probs = master_counts / master_counts.sum()

    return master_probs


def calculate_2d_probabilities_with_nans(x_data, y_data, x_bins, y_bins):
    x_nan_mask = np.isnan(x_data)
    y_nan_mask = np.isnan(y_data)

    both_nan_mask = x_nan_mask & y_nan_mask

    clean_x_data = x_data[~both_nan_mask]
    clean_y_data = y_data[~both_nan_mask]

    counts, _, _ = np.histogram2d(clean_x_data, clean_y_data, [x_bins, y_bins])

    x_counts_when_x_is_nan, _ = np.histogram(x_data[y_nan_mask], x_bins)
    y_counts_when_x_is_nan, _ = np.histogram(y_data[x_nan_mask], y_bins)

    both_nan_counts = both_nan_mask.sum()

    master_counts = np.empty((counts.shape[0] + 1, counts.shape[1] + 1), dtype=int)
    master_counts[:-1, :-1] = counts
    master_counts[-1, :-1] = y_counts_when_x_is_nan
    master_counts[:-1, -1] = x_counts_when_x_is_nan
    master_counts[-1, -1] = both_nan_counts

    master_probs = master_counts / master_counts.sum()

    return master_probs


def make_clean_bins_from_data(data):
    nan_mask = np.isnan(data)
    clean_data = data[~nan_mask]
    N = len(clean_data)

    if N == 0:
        min_ = 0
        max_ = 0
        mean = 0
        std = 1
        N = 1
    elif N == 1:
        min_ = clean_data.min()
        max_ = clean_data.max()
        mean = clean_data.mean()
        std = 1
    else:
        min_ = clean_data.min()
        max_ = clean_data.max()
        mean = clean_data.mean()
        std = clean_data.std(ddof=1)

    # scott's bin width rule
    bin_width = 3.49 * std / (N**1 / 3)
    bin_edges = construct_bin_edges(mean, min_, max_, bin_width)
    return bin_edges


def construct_bin_edges(center, min_, max_, bin_width):
    assert bin_width > 0.0
    assert max_ >= min_
    assert max_ >= center
    assert min_ <= center

    num_down_steps = np.ceil((center - bin_width / 2 - min_) / bin_width)
    num_up_steps = np.ceil((max_ - (center + bin_width / 2)) / bin_width)

    low_end = center - bin_width / 2 - num_down_steps * bin_width
    high_end = center + bin_width / 2 + num_up_steps * bin_width

    num_edges = int(num_down_steps + num_up_steps + 2)
    bin_edges = np.linspace(low_end, high_end, num_edges)

    return bin_edges


def mutual_information(x_probs, y_probs, xy_probs, eps: float = 1e-10):
    independent_probs_matrix = np.outer(x_probs, y_probs)
    independent_probs_matrix[independent_probs_matrix == 0] = 1
    ratio_matrix = np.divide(xy_probs, independent_probs_matrix)
    ratio_matrix[ratio_matrix == 0] = eps
    log_matrix = np.log2(ratio_matrix)
    entropies = np.multiply(xy_probs, log_matrix)

    return entropies.sum()



import pytest

from io import BytesIO, SEEK_CUR, SEEK_END, SEEK_SET
from mehr.stream import StreamProgressWrapper

LINE_ODD = b"QWERTZUIOP\n"
LINE_EVEN = b"ASDFGHJKLYXCVBNM\n"

NUM_LINE_REPEAT = 10_000

NUM_SMALL_READ = 8

FIRST_SEEK = 40
SECOND_SEEK = 20
THIRD_SEEK = -50

EXPECTED_SIZE = 420

BAD_SEEK_MODE_VALUES = [-1, -42, 3]
BAD_SEEK_MODE_TYPES = [1.0, "2", b"1"]


def bytes_stream():
    stream = BytesIO()
    for _ in range(NUM_LINE_REPEAT):
        stream.write(LINE_ODD)
        stream.write(LINE_EVEN)
    stream.seek(0)
    final_pos = NUM_LINE_REPEAT * (len(LINE_ODD) + len(LINE_EVEN))

    return stream, final_pos


def test_read():
    stream, _ = bytes_stream()
    with StreamProgressWrapper(stream) as f:
        data = f.read(NUM_SMALL_READ)

    assert data == LINE_ODD[:NUM_SMALL_READ]


def test_seek_tell():
    stream, final_pos = bytes_stream()

    with StreamProgressWrapper(stream) as f:
        assert f.tell() == 0
        f.seek(FIRST_SEEK)
        assert f.tell() == FIRST_SEEK
        f.seek(SECOND_SEEK, SEEK_CUR)
        assert f.tell() == FIRST_SEEK + SECOND_SEEK
        f.seek(THIRD_SEEK, SEEK_END)
        assert f.tell() == final_pos + THIRD_SEEK


def test_bad_seek_mode_values():
    for bad_mode in BAD_SEEK_MODE_VALUES:
        stream, _ = bytes_stream()
        with StreamProgressWrapper(stream) as f:
            with pytest.raises(ValueError):
                f.seek(FIRST_SEEK, bad_mode)


def test_bad_seek_mode_types():
    for bad_mode in BAD_SEEK_MODE_TYPES:
        stream, _ = bytes_stream()
        with StreamProgressWrapper(stream) as f:
            with pytest.raises(TypeError):
                f.seek(FIRST_SEEK, bad_mode)
        assert f.closed


def test_neg_seek_position():
    stream, final_pos = bytes_stream()
    with StreamProgressWrapper(stream) as f:
        with pytest.raises(ValueError):
            f.seek(-1, SEEK_SET)


def test_far_seek():
    stream, final_pos = bytes_stream()
    with StreamProgressWrapper(stream) as f:
        pos = f.seek(final_pos * 100, SEEK_SET)
        assert pos == final_pos * 100
        rest_data = f.read(NUM_SMALL_READ)
        assert rest_data == b""


def test_close():
    stream, final_pos = bytes_stream()
    with StreamProgressWrapper(stream) as f:
        f.close()

    assert f.closed
    assert stream.closed


def test_guessed_size():
    stream, final_pos = bytes_stream()
    with StreamProgressWrapper(stream, guess_size=True) as f:
        assert f.size == final_pos


def test_expected_size():
    stream, final_pos = bytes_stream()
    with StreamProgressWrapper(stream, expected_size=EXPECTED_SIZE) as f:
        assert f.size == EXPECTED_SIZE


def test_readline():
    stream, final_pos = bytes_stream()
    with StreamProgressWrapper(stream) as f:
        line_odd = f.readline()
        assert line_odd == LINE_ODD
        line_even = f.readline()
        assert line_even == LINE_EVEN
        line_odd2 = f.readline()
        assert line_odd2 == LINE_ODD
        line_even2 = f.readline()
        assert line_even2 == LINE_EVEN


def test_readlines():
    stream, final_pos = bytes_stream()
    with StreamProgressWrapper(stream) as f:
        lines = f.readlines(2 * len(LINE_ODD) + 2 * len(LINE_EVEN))
        assert len(lines) == 4
        assert lines[0] == LINE_ODD
        assert lines[1] == LINE_EVEN
        assert lines[2] == LINE_ODD
        assert lines[3] == LINE_EVEN



import pytest

from mehr.analysis.mifs import binifier


@pytest.mark.parametrize("center,min_,max_,bin_width,low,high", [
    (0.5, 0, 1, 1, 0, 1),
    (0.5, 0.1, 0.9, 1, 0, 1),
    (0.5, -0.1, 1.1, 1, -1, 2),
    (3.14, 1, 7, 0.5, 0.89, 7.39)
])
def test_binifier(center, min_, max_, bin_width, low, high):
    bin_edges = binifier(center, min_, max_, bin_width)
    
    bin_edges[0] == pytest.approx(low)
    bin_edges[-1] == pytest.approx(high)


import io
from types import TracebackType
from typing import Any, List, Optional, Type, Union
from typing_extensions import Self
from tqdm import tqdm
from io import SEEK_CUR, SEEK_END, SEEK_SET

SeekMode = Union[type(SEEK_SET), type(SEEK_CUR), type(SEEK_END)]


class StreamProgressWrapper(io.IOBase):
    def __init__(
        self, stream: io.IOBase, *, desc: str = "", expected_size=None, guess_size=True
    ):
        self.stream = stream
        self.size = None
        self.desc = desc

        if expected_size is not None:
            self.size = expected_size
        elif guess_size and self.stream.seekable():
            pos = stream.tell()
            stream.seek(0, SEEK_END)
            self.size = stream.tell()
            stream.seek(pos, SEEK_SET)
        else:
            pass

        self.progress_bar = tqdm(total=self.size, desc=desc)

    def __enter__(self) -> Self:
        self.stream.__enter__()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Any:
        return self.stream.__exit__(exc_type, exc_val, exc_tb)

    def close(self) -> None:
        self.stream.close()

    @property
    def closed(self) -> bool:
        return self.stream.closed

    def read(self, hint=-1):
        data = self.stream.read(hint)
        self.progress_bar.update(len(data))
        return data

    def readline(self, size: Optional[int] = None):
        data = self.stream.readline(size)
        self.progress_bar.update(len(data))
        return data

    def readlines(self, hint: Optional[int] = None) -> List:
        lines = self.stream.readlines(hint)
        num_data = sum([len(line) for line in lines])
        self.progress_bar.update(num_data)
        return lines

    def seek(self, offset: int, whence: SeekMode = SEEK_SET) -> int:
        pos = self.stream.seek(offset, whence)
        self.progress_bar.n = pos
        return pos

    def tell(self) -> int:
        return self.stream.tell()



from setuptools import setup

VERSION = "0.0.2"

setup(
    name="mehr",
    version=f"{VERSION}",
    author="Moraxno",
    author_email="robertbock.98@googlemail.com",
    packages=["mehr"],
    scripts=[],
    url="http://pypi.python.org/pypi/mehr/",
    license="LICENSE.txt",
    description="Helpers",
    long_description=open("README.md").read(),
    download_url=f"https://github.com/Moraxno/py-mehr/archive/refs/tags/v{VERSION}.tar.gz",
    install_requires=[
        "pytest",
        "tqdm"
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
)
